// $Id: bpla_kernel.cpp 293 2009-01-16 11:18:31Z satoken $

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif
#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <cmath>
#include <boost/multi_array.hpp>
#include "bpla_kernel.h"

template < class ValueType, class Data >
struct LAScore
{
  LAScore(const boost::multi_array<ValueType,2>& score_table)
    : score_table_(score_table)
  { }

  inline
  ValueType
  operator()(const Data& x, const Data& y, uint i, uint j) const
  {
    ValueType v = 0.0;
    float n = 0;
    for (uint k=0; k!=N_RNA; ++k) {
      if (x.seq[i][k]==0) continue;
      for (uint l=0; l!=N_RNA; ++l) {
	if (y.seq[j][l]==0) continue;
	n += x.seq[i][k]*y.seq[j][l];
	v += score_table_[k][l]*x.seq[i][k]*y.seq[j][l];
      }
    }
    return n==0 ? 0.0 : v/n;
  }

  const boost::multi_array<ValueType,2>& score_table_;
};

template < class ValueType, class Data >
struct BPLAScore : public LAScore<ValueType,Data>
{
  BPLAScore(const boost::multi_array<ValueType,2>& score_table, ValueType alpha,
	    const Data& x, const Data& y)
    : LAScore<ValueType,Data>(score_table), alpha_(alpha) 
  { }

  inline
  ValueType
  operator()(const Data& x, const Data& y, uint i, uint j) const
  {
    return alpha_ * (x.p_right[i]*y.p_right[j] + x.p_left[i]*y.p_left[j])
      + x.p_unpair[i]*y.p_unpair[j] * LAScore<ValueType,Data>::operator()(x,y,i,j);
  }

  ValueType alpha_;
};

template < class Data, class ValueType, class Score >
inline
ValueType
local_alignment(const Data& x, const Data& y,
		ValueType beta,	ValueType gap, ValueType ext,
		Score score)
{
  ValueType beta_gap = exp(beta*gap);
  ValueType beta_ext = exp(beta*ext);
  typedef boost::multi_array<ValueType,2> dp_type;
  dp_type M(boost::extents[x.size()+1][y.size()+1]);
  dp_type X(boost::extents[x.size()+1][y.size()+1]);
  dp_type Y(boost::extents[x.size()+1][y.size()+1]);
  dp_type X2(boost::extents[x.size()+1][y.size()+1]);
  dp_type Y2(boost::extents[x.size()+1][y.size()+1]);

  //initialize  
  for (uint i=0; i!=x.size()+1; ++i) {
    M[i][0]=0;
    X[i][0]=0;
    Y[i][0]=0;
    X2[i][0]=0;
    Y2[i][0]=0;
  }

  for (uint j=0; j!=y.size()+1; ++j) {
    M[0][j]=0;
    X[0][j]=0;
    Y[0][j]=0;
    X2[0][j]=0;
    Y2[0][j]=0;
  }

  //calculate
  for (uint i=1; i != x.size()+1; ++i) {
    for (uint j=1; j != y.size()+1; ++j) {
      M[i][j] = exp(beta * score(x, y, i-1, j-1))
	* (1 + X[i-1][j-1] + Y[i-1][j-1] + M[i-1][j-1]);
      X[i][j] = beta_gap * M[i-1][j] 
	+ beta_ext * X[i-1][j] ;
      
      Y[i][j] = beta_gap * (M[i][j-1] + X[i][j-1]) 
	+ beta_ext * Y[i][j-1] ;

      X2[i][j]= M[i-1][j] + X2[i-1][j];
      Y2[i][j]= M[i][j-1] + X2[i][j-1] + Y2[i][j-1];
    }
  }

  return 1+X2[x.size()][y.size()]
    + Y2[x.size()][y.size()] + M[x.size()][y.size()];
}

/*
template < class Data, class ValueType, class Score >
inline
ValueType
local_alignment(const Data& x, const Data& y,
		ValueType beta,	ValueType gap, ValueType ext,
		Score score)
{
  ValueType beta_gap = exp(beta*gap);
  ValueType beta_ext = exp(beta*ext);
  typedef boost::multi_array<ValueType,2> dp_type;
  dp_type N(boost::extents[x.size()+1][y.size()+1]);
  dp_type N2(boost::extents[x.size()+1][y.size()+1]);
  dp_type M(boost::extents[x.size()+1][y.size()+1]);
  dp_type X(boost::extents[x.size()+1][y.size()+1]);
  dp_type Y(boost::extents[x.size()+1][y.size()+1]);
  dp_type X0(boost::extents[x.size()+1][y.size()+1]);
  dp_type Y0(boost::extents[x.size()+1][y.size()+1]);
  dp_type X2(boost::extents[x.size()+1][y.size()+1]);
  dp_type Y2(boost::extents[x.size()+1][y.size()+1]);

  //initialize  
  for (uint i=0; i!=x.size()+1; ++i) {
    N[i][0]=0;
    N2[i][0]=0;
    M[i][0]=0;
    X[i][0]=0;
    Y[i][0]=0;
    X0[i][0]=0;
    Y0[i][0]=0;
    X2[i][0]=0;
    Y2[i][0]=0;
  }

  for (uint j=0; j!=y.size()+1; ++j) {
    N[0][j]=0;
    N2[0][j]=0;
    M[0][j]=0;
    X[0][j]=0;
    Y[0][j]=0;
    X0[0][j]=0;
    Y0[0][j]=0;
    X2[0][j]=0;
    Y2[0][j]=0;
  }

  //calculate
  for (uint i=1; i != x.size()+1; ++i) {
    for (uint j=1; j != y.size()+1; ++j) {
      N[i][j] = exp(beta * score(x, y, i-1, j-1))
	* (1 + X0[i-1][j-1] + Y0[i-1][j-1]);

      N2[i][j] = exp(beta * score(x, y, i-1, j-1)) 
	* N[i-1][j-1];

      M[i][j] = exp(beta * score(x, y, i-1, j-1)) 
	* (N2[i-1][j-1] + M[i-1][j-1] + X[i-1][j-1] + Y[i-1][j-1]);

      X[i][j] = beta_gap * M[i-1][j] 
	+ beta_ext * X[i-1][j] ;
      
      Y[i][j] = beta_gap * (M[i][j-1] + X[i][j-1]) 
	+ beta_ext * Y[i][j-1] ;

      X0[i][j] = beta_gap * (N[i-1][j] + N2[i-1][j]) 
	+ beta_ext * X0[i-1][j] ;
      
      Y0[i][j] = beta_gap * (N[i][j-1] + N2[i][j-1] + X0[i][j-1]) 
	+ beta_ext * Y0[i][j-1] ;

      X2[i][j]= M[i-1][j] + X2[i-1][j];
      Y2[i][j]= M[i][j-1] + X2[i][j-1] + Y2[i][j-1];
    }
  }

  return 1+X2[x.size()][y.size()]
    + Y2[x.size()][y.size()] + M[x.size()][y.size()];
}
*/

template < class ValueType, class Data >
ValueType
BPLAKernel<ValueType,Data>::
operator()(const Data& x, const Data& y) const
{
  if (noBP_)
    return local_alignment(x, y, beta_, gap_, ext_, LAScore<ValueType,Data>(score_table_));
  else
    return local_alignment(x, y, beta_, gap_, ext_, BPLAScore<ValueType,Data>(score_table_, alpha_, x, y));
}


/*
enum { M=0, IX=1, IY=2, LX=3, LY=4, RX=5, RY=6, N=7 };

template < class ValueType, class Data, class Table >
ValueType
BPLA_Forward(const Data& x, const Data& y,
             const boost::multi_array<ValueType,2>& score_table,
	     const std::vector<double>& param, Table& T)
{
  assert(T.size()==N);
  LAScore<ValueType, Data> la_score(score_table);
  ValueType alpha = param[0];
  ValueType beta = param[1];
  ValueType gap = param[2];
  ValueType ext = param[3];
  ValueType beta_gap = exp(beta*gap);
  ValueType beta_ext = exp(beta*ext);

  std::fill(T.data(), T.data()+T.num_elements(), 0.0);
  T[M][0][0]=1;
  T[LX][0][0]=1;
  T[LY][0][0]=1;
  
  for (uint i=1; i!=x.size()+1; ++i) {
    T[LX][i][0]+=T[LX][i-1][0];
  }

  for (uint j=1; j!=y.size()+1; ++j) {
    T[LY][0][j]+=T[LY][0][j-1];
  }

  for (uint i=1; i!= x.size()+1; ++i) {
    for (uint j=1; j!= y.size()+1; ++j) {
      ValueType s =
	alpha * (x.p_right[i-1]*y.p_right[j-1] + x.p_left[i-1] * y.p_left[j-1])
	+ x.p_unpair[i-1]*y.p_unpair[j-1] * la_score(x, y, i-1, j-1);
      ValueType beta_s = exp(beta * s);
      T[M][i][j] += beta_s*T[M][i-1][j-1];
      T[M][i][j] += beta_s*T[IX][i-1][j-1];
      T[M][i][j] += beta_s*T[IY][i-1][j-1];
      T[M][i][j] += beta_s*T[LX][i-1][j-1];
      T[M][i][j] += beta_s*T[LY][i-1][j-1];

      T[IX][i][j] += beta_gap*T[M][i-1][j];
      T[IX][i][j] += beta_ext*T[IX][i-1][j];

      T[IY][i][j] += beta_gap*T[M][i][j-1];
      T[IY][i][j] += beta_gap*T[IX][i][j-1];
      T[IY][i][j] += beta_ext*T[IY][i][j-1];

      T[LX][i][j] += T[LX][i-1][0];

      T[LY][i][j] += T[LX][i][j-1];
      T[LY][i][j] += T[LY][i][j-1];

      T[RX][i][j] += T[M][i-1][j];
      T[RX][i][j] += T[RX][i-1][j];

      T[RY][i][j] += T[M][i][j-1];
      T[RY][i][j] += T[RX][i][j-1];
      T[RY][i][j] += T[RY][i][j-1];
    }
  }

  return 1+T[M][x.size()][y.size()]
    +T[RX][x.size()][y.size()]+T[RY][x.size()][y.size()];
}

template < class ValueType, class Data, class Table >
ValueType
BPLA_Backward(const Data& x, const Data& y,
              const boost::multi_array<ValueType,2>& score_table,
	      const std::vector<double>& param, Table& T)
{
  assert(T.size()==N);
  LAScore<ValueType, Data> la_score(score_table);
  ValueType alpha = param[0];
  ValueType beta = param[1];
  ValueType gap = param[2];
  ValueType ext = param[3];
  ValueType beta_gap = exp(beta*gap);
  ValueType beta_ext = exp(beta*ext);

  std::fill(T.data(), T.data()+T.num_elements(), 0.0);
  T[M][x.size()][y.size()]=1;
  T[RX][x.size()][y.size()]=1;
  T[RY][x.size()][y.size()]=1;
  
  for (uint i=x.size(); i!=0; --i) {
    for (uint j=y.size(); j!=0; --j) {
      ValueType s =
	alpha * (x.p_right[i-1]*y.p_right[j-1] + x.p_left[i-1] * y.p_left[j-1])
	+ x.p_unpair[i-1]*y.p_unpair[j-1] * la_score(x, y, i-1, j-1);
      ValueType beta_s = exp(beta * s);
      T[M][i-1][j-1] += beta_s*T[M][i][j] ;
      T[IX][i-1][j-1] += beta_s*T[M][i][j];
      T[IY][i-1][j-1] += beta_s*T[M][i][j];
      T[LX][i-1][j-1] += beta_s*T[M][i][j];
      T[LY][i-1][j-1] += beta_s*T[M][i][j];

      T[M][i-1][j] += beta_gap*T[IX][i][j];
      T[IX][i-1][j] += beta_ext*T[IX][i][j];

      T[M][i][j-1] += beta_gap*T[IY][i][j];
      T[IX][i][j-1] += beta_gap*T[IY][i][j];
      T[IY][i][j-1] += beta_ext*T[IY][i][j];

      T[LX][i-1][0] += T[LX][i][j];

      T[LX][i][j-1] += T[LY][i][j];
      T[LY][i][j-1] += T[LY][i][j];

      T[M][i-1][j] += T[RX][i][j];
      T[RX][i-1][j] += T[RX][i][j];

      T[M][i][j-1] += T[RY][i][j];
      T[RX][i][j-1] += T[RY][i][j];
      T[RY][i][j-1] += T[RY][i][j];
    }
  }

  for (uint i=x.size(); i!=0; --i) {
    T[LX][i-1][0]+=T[LX][i][0];
  }

  for (uint j=y.size(); j!=0; --j) {
    T[LY][0][j-1]+=T[LY][0][j];
  }

  return 1+T[M][0][0]+T[LX][0][0]+T[LY][0][0];
}

template < class V >
inline
void
update_alpha_beta(double& d_a, double& d_b, V a, V b, V w_pair, V w_unpair, V v)
{
  d_a += b*w_pair*v;
  d_b += (a*w_pair+w_unpair)*v;
}

template < class V >
inline
void
update_beta_gap_ext(double& d_b, double& d_ge, V b, V ge, V v)
{
  d_b  += ge*v;
  d_ge +=  b*v;
}
  
template < class ValueType, class Data, class Table >
ValueType
BPLA_ForwardBackword(const Data& x, const Data& y,
                     const boost::multi_array<ValueType,2>& score_table,
		     const std::vector<double>& param,
		     const Table& F, const Table& B,
		     std::vector<double>& d)
{
  assert(F.size()==N);
  assert(B.size()==N);
  LAScore<ValueType, Data> la_score(score_table);
  ValueType alpha = param[0];
  ValueType beta = param[1];
  ValueType gap = param[2];
  ValueType ext = param[3];
  ValueType beta_gap = exp(beta*gap);
  ValueType beta_ext = exp(beta*ext);

  assert(d.size()==param.size());
  double& d_alpha = d[0];
  double& d_beta = d[1];
  double& d_gap = d[2];
  double& d_ext = d[3];

  for (uint i=1; i!= x.size()+1; ++i) {
    for (uint j=1; j!= y.size()+1; ++j) {
      ValueType w_pair=x.p_right[i-1]*y.p_right[j-1] + x.p_left[i-1] * y.p_left[j-1];
      ValueType w_unpair=x.p_unpair[i-1]*y.p_unpair[j-1] * la_score(x, y, i-1, j-1);
      ValueType beta_s = exp(beta * (alpha * w_pair + w_unpair));
      
      update_alpha_beta(d_alpha, d_beta, alpha, beta, w_pair, w_unpair,
			F[M][i-1][j-1]*beta_s*B[M][i][j]);
      update_alpha_beta(d_alpha, d_beta, alpha, beta, w_pair, w_unpair,
			F[IX][i-1][j-1]*beta_s*B[M][i][j]);
      update_alpha_beta(d_alpha, d_beta, alpha, beta, w_pair, w_unpair,
			F[IY][i-1][j-1]*beta_s*B[M][i][j]);
      update_alpha_beta(d_alpha, d_beta, alpha, beta, w_pair, w_unpair,
			F[LX][i-1][j-1]*beta_s*B[M][i][j]);
      update_alpha_beta(d_alpha, d_beta, alpha, beta, w_pair, w_unpair,
			F[LY][i-1][j-1]*beta_s*B[M][i][j]);

      update_beta_gap_ext(d_beta, d_gap, beta, gap, 
			  F[M][i-1][j]*beta_gap*B[IX][i][j]);
      update_beta_gap_ext(d_beta, d_ext, beta, ext, 
			  F[IX][i-1][j]*beta_ext*B[IX][i][j]);

      update_beta_gap_ext(d_beta, d_gap, beta, gap, 
			  F[M][i][j-1]*beta_gap*B[IY][i][j]);
      update_beta_gap_ext(d_beta, d_gap, beta, gap, 
			  F[IX][i][j-1]*beta_gap*B[IY][i][j]);
      update_beta_gap_ext(d_beta, d_ext, beta, ext, 
			  F[IY][i][j-1]*beta_ext*B[IY][i][j]);
    }
  }

  return 1+F[M][x.size()][y.size()]
    +F[RX][x.size()][y.size()]+F[RY][x.size()][y.size()];
}

// static
template < class ValueType, class Data >
ValueType
BPLAKernel<ValueType,Data>::
compute_gradients(const Data& x, const Data& y,
                  const boost::multi_array<value_type,2>& score_table,
		  const std::vector<double>& param,
		  std::vector<double>& d)
{
  std::fill(d.begin(), d.end(), 0.0);
  typedef boost::multi_array<ValueType,3> dp_type;
  dp_type F(boost::extents[N][x.size()+1][y.size()+1]);
  dp_type B(boost::extents[N][x.size()+1][y.size()+1]);
  BPLA_Forward<ValueType,Data,dp_type>(x, y, score_table, param, F);
  BPLA_Backward<ValueType,Data,dp_type>(x, y, score_table, param, B);
  return BPLA_ForwardBackword<ValueType,Data,dp_type>(x, y, score_table, param, F, B, d);
}
*/


// static
template < class ValueType, class Data >
ValueType
BPLAKernel<ValueType,Data>::
compute_gradients(const Data& x, const Data& y,
                  const boost::multi_array<value_type,2>& score_table,
		  const std::vector<double>& param,
		  std::vector<double>& d)
{
  LAScore<ValueType, Data> la_score(score_table);
  ValueType len = (double) (x.size()+y.size()) / 2;
  ValueType alpha = param[0];
  ValueType beta = param[1];
  ValueType gap = param[2];
  ValueType ext = param[3];
  ValueType beta_gap = exp(beta*gap);
  ValueType beta_ext = exp(beta*ext);
  ValueType beta_gap2 = gap * beta_gap;
  ValueType beta_ext2 = ext * beta_ext;
  ValueType beta2_gap = beta * beta_gap;
  ValueType beta2_ext = beta * beta_ext;

  typedef boost::multi_array<ValueType,2> dp_type;

  // first, perform the DP of BPLA kernels itself
  dp_type M(boost::extents[x.size()+1][y.size()+1]);
  dp_type X(boost::extents[x.size()+1][y.size()+1]);
  dp_type Y(boost::extents[x.size()+1][y.size()+1]);
  dp_type X2(boost::extents[x.size()+1][y.size()+1]);
  dp_type Y2(boost::extents[x.size()+1][y.size()+1]);

  for (uint i=0; i!=x.size()+1; ++i) {
    M[i][0]=0;
    X[i][0]=0;
    Y[i][0]=0;
    X2[i][0]=0;
    Y2[i][0]=0;
  }
  for (uint j=0; j!=y.size()+1; ++j) {
    M[0][j]=0;
    X[0][j]=0;
    Y[0][j]=0;
    X2[0][j]=0;
    Y2[0][j]=0;
  }

  for (uint i=1; i != x.size()+1; ++i) {
    for (uint j=1; j != y.size()+1; ++j) {
      double str = x.p_right[i-1]*y.p_right[j-1] + x.p_left[i-1]*y.p_left[j-1];
      double seq = x.p_unpair[i-1]*y.p_unpair[j-1] * la_score(x,y,i-1,j-1);
      double score = alpha * str + seq;

      M[i][j] = exp(beta * score)
	* (1 + X[i-1][j-1] + Y[i-1][j-1] + M[i-1][j-1]);

      X[i][j] = beta_gap * M[i-1][j] 
	+ beta_ext * X[i-1][j] ;
      Y[i][j] = beta_gap * (M[i][j-1] + X[i][j-1]) 
	+ beta_ext * Y[i][j-1] ;

      X2[i][j]= M[i-1][j] + X2[i-1][j];
      Y2[i][j]= M[i][j-1] + X2[i][j-1] + Y2[i][j-1];
    }
  }

  // then, perform the DP of gradient calculation
  dp_type PM(boost::extents[x.size()+1][y.size()+1]);
  dp_type PX(boost::extents[x.size()+1][y.size()+1]);
  dp_type PY(boost::extents[x.size()+1][y.size()+1]);
  dp_type PX2(boost::extents[x.size()+1][y.size()+1]);
  dp_type PY2(boost::extents[x.size()+1][y.size()+1]);
  std::fill(d.begin(), d.end(), 0.0);

  for (uint i=0; i!=x.size()+1; ++i) {
    PM[i][0]=0;
    PX[i][0]=0;
    PY[i][0]=0;
    PX2[i][0]=0;
    PY2[i][0]=0;
  }
  for (uint j=0; j!=y.size()+1; ++j) {
    PM[0][j]=0;
    PX[0][j]=0;
    PY[0][j]=0;
    PX2[0][j]=0;
    PY2[0][j]=0;
  }

  // alpha
  for (uint i=1; i != x.size()+1; ++i) {
    for (uint j=1; j != y.size()+1; ++j) {
      double str = x.p_right[i-1]*y.p_right[j-1] + x.p_left[i-1]*y.p_left[j-1];
      double seq = x.p_unpair[i-1]*y.p_unpair[j-1] * la_score(x,y,i-1,j-1);
      double score = alpha * str + seq;

      PM[i][j] = beta * M[i][j] * str 
	+ exp(beta*score) * (PX[i-1][j-1] + PY[i-1][j-1] + PM[i-1][j-1]);

      PX[i][j] = beta_gap * PM[i-1][j] 
	+ beta_ext * PX[i-1][j] ;
      
      PY[i][j] = beta_gap * (PM[i][j-1] + PX[i][j-1]) 
	+ beta_ext * PY[i][j-1] ;

      PX2[i][j]= PM[i-1][j] + PX2[i-1][j];
      PY2[i][j]= PM[i][j-1] + PX2[i][j-1] + PY2[i][j-1];
    }
  }
  d[0] = PX2[x.size()][y.size()] + PY2[x.size()][y.size()] + PM[x.size()][y.size()];

  // beta
  for (uint i=1; i != x.size()+1; ++i) {
    for (uint j=1; j != y.size()+1; ++j) {
      double str = x.p_right[i-1]*y.p_right[j-1] + x.p_left[i-1]*y.p_left[j-1];
      double seq = x.p_unpair[i-1]*y.p_unpair[j-1] * la_score(x,y,i-1,j-1);
      double score = alpha * str + seq;

      PM[i][j] = M[i][j] * score 
	+ exp(beta*score) * (PX[i-1][j-1] + PY[i-1][j-1] + PM[i-1][j-1]);

      PX[i][j] = beta_gap2 * M[i-1][j] + beta_ext2 * X[i-1][j]
	+ beta_gap * PM[i-1][j] + beta_ext * PX[i-1][j];
      
      PY[i][j] = beta_gap2 * (M[i][j-1] + X[i][j-1]) + beta_ext2 * Y[i][j-1]
	+ beta_gap * (PM[i][j-1] + PX[i][j-1]) + beta_ext * PY[i][j-1];

      PX2[i][j]= PM[i-1][j] + PX2[i-1][j];
      PY2[i][j]= PM[i][j-1] + PX2[i][j-1] + PY2[i][j-1];
    }
  }
  d[1] = PX2[x.size()][y.size()] + PY2[x.size()][y.size()] + PM[x.size()][y.size()];

  // gap
  for (uint i=1; i != x.size()+1; ++i) {
    for (uint j=1; j != y.size()+1; ++j) {
      double str = x.p_right[i-1]*y.p_right[j-1] + x.p_left[i-1]*y.p_left[j-1];
      double seq = x.p_unpair[i-1]*y.p_unpair[j-1] * la_score(x,y,i-1,j-1);
      double score = alpha * str + seq;

      PM[i][j] = exp(beta*score) * (PX[i-1][j-1] + PY[i-1][j-1] + PM[i-1][j-1]);

      PX[i][j] = beta2_gap * M[i-1][j]
	+ beta_gap * PM[i-1][j] + beta_ext * PX[i-1][j];
      
      PY[i][j] = beta2_gap * (M[i][j-1] + X[i][j-1]) 
	+ beta_gap * (PM[i][j-1] + PX[i][j-1]) + beta_ext * PY[i][j-1];

      PX2[i][j]= PM[i-1][j] + PX2[i-1][j];
      PY2[i][j]= PM[i][j-1] + PX2[i][j-1] + PY2[i][j-1];
    }
  }
  d[2] = PX2[x.size()][y.size()] + PY2[x.size()][y.size()] + PM[x.size()][y.size()];

  // ext
  for (uint i=1; i != x.size()+1; ++i) {
    for (uint j=1; j != y.size()+1; ++j) {
      double str = x.p_right[i-1]*y.p_right[j-1] + x.p_left[i-1]*y.p_left[j-1];
      double seq = x.p_unpair[i-1]*y.p_unpair[j-1] * la_score(x,y,i-1,j-1);
      double score = alpha * str + seq;

      PM[i][j] = exp(beta*score) * (PX[i-1][j-1] + PY[i-1][j-1] + PM[i-1][j-1]);

      PX[i][j] = beta2_ext * X[i-1][j]
	+ beta_gap * PM[i-1][j] + beta_ext * PX[i-1][j];
      
      PY[i][j] = beta2_ext * Y[i][j-1]
	+ beta_gap * (PM[i][j-1] + PX[i][j-1]) + beta_ext * PY[i][j-1];

      PX2[i][j]= PM[i-1][j] + PX2[i-1][j];
      PY2[i][j]= PM[i][j-1] + PX2[i][j-1] + PY2[i][j-1];
    }
  }
  d[3] = PX2[x.size()][y.size()] + PY2[x.size()][y.size()] + PM[x.size()][y.size()];

  return 1+X2[x.size()][y.size()]
    + Y2[x.size()][y.size()] + M[x.size()][y.size()];
}


// instatiation
template class BPLAKernel<double,MData>;
